@@ -1,5 +1,6 @@
+# pylint: disable=too-many-lines,R0903,W0142,R0913,C0302
 # -*- coding: utf-8 -*-
-# pylint: disable=R0903,W0142
+
 
 """ This module contains all the data structures for the information objects
 such as music tracks or genres
@@ -7,11 +8,8 @@ such as music tracks or genres
 """
 
 from __future__ import unicode_literals
-try:
-    import xml.etree.cElementTree as XML
-except ImportError:
-    import xml.etree.ElementTree as XML
 
+from .xml import XML
 from .exceptions import CannotCreateDIDLMetadata
 from .utils import really_unicode, camel_to_underscore
 
@@ -34,12 +32,13 @@ def get_ml_item(xml):
     return cls.from_xml(xml=xml)
 
 
-def get_ms_item(xml, service):
+def get_ms_item(xml, service, parent_id):
     """Return the music service item that corresponds to xml. The class is
     identified by getting the type from the 'itemType' tag
     """
-    cls = MS_TYPE_TO_CLASS[xml.findtext(ns_tag('ms', 'itemType'))]
-    return cls.from_xml(xml, service)
+    cls = MS_TYPE_TO_CLASS.get(xml.findtext(ns_tag('ms', 'itemType')))
+    out = cls.from_xml(xml, service, parent_id)
+    return out
 
 
 def tags_with_text(xml, tags=None):
@@ -54,7 +53,7 @@ def tags_with_text(xml, tags=None):
         elif len(element) > 0:
             tags_with_text(element, tags)
         else:
-            message = 'Unknown XML structure: {}'.format(element)
+            message = 'Unknown XML structure: {0}'.format(element)
             raise ValueError(message)
     return tags
 
@@ -70,6 +69,11 @@ class MusicInfoItem(object):
         """Return the equals comparison result to another ``playable_item``."""
         return self.content == playable_item.content
 
+    def __ne__(self, playable_item):
+        """Return the not equals comparison result to another ``playable_item``
+        """
+        return self.content != playable_item.content
+
     def __repr__(self):
         """Return the repr value for the item.
 
@@ -206,7 +210,7 @@ class MusicLibraryItem(MusicInfoItem):
     @property
     def to_dict(self):
         """Get the dict representation of the instance."""
-        return self.content
+        return self.content.copy()
 
     @property
     def item_id(self):  # pylint: disable=C0103
@@ -610,6 +614,32 @@ class MLPlaylist(MusicLibraryItem):
         return out
 
 
+class MLSonosPlaylist(MusicLibraryItem):
+    """ Class that represents a sonos playlist.
+
+    :ivar parent_id: The parent ID for the MLSonosPlaylist is 'SQ:'
+    :ivar _translation: The dictionary-key-to-xml-tag-and-namespace-
+        translation used when instantiating MLSonosPlaylist from
+        XML is inherited from :py:class:`.MusicLibraryItem`.
+
+    """
+
+    parent_id = 'SQ:'
+
+    def __init__(self, uri, title,
+                 item_class='object.container.playlistContainer'):
+        """ Instantiate the MLSonosPlaylist item by passing the arguments to
+        the super class :py:meth:`.MusicLibraryItem.__init__`.
+
+        :param uri: The URI for the playlist
+        :param title: The title of the playlist
+        :param item_class: The UPnP class for the playlist. The default value
+            is: ``object.container.playlistContainer``
+
+        """
+        MusicLibraryItem.__init__(self, uri, title, item_class)
+
+
 class MLShare(MusicLibraryItem):
     """Class that represents a music library share.
 
@@ -635,6 +665,63 @@ class MLShare(MusicLibraryItem):
         MusicLibraryItem.__init__(self, uri, title, item_class)
 
 
+###############################################################################
+# MUSIC LIBRARY                                                               #
+###############################################################################
+class URI(MusicInfoItem):
+    """General purpose class that represents a audio file URI"""
+
+    valid_fields = ['uri']
+    # Note, keep this list sorted with the args to __init__
+    required_fields = ['uri']
+
+    def __init__(self, uri):
+        """Init method for the URI object"""
+        super(URI, self).__init__()
+        self.content.update({'uri': uri})
+
+    @classmethod
+    def from_dict(cls, dict_in):
+        """Initialize a URI item from a dict"""
+        # Copy before we modify
+        kwargs = dict_in.copy()
+        # Check input
+        for key in kwargs.keys():
+            if key not in cls.valid_fields:
+                message = 'The key \'{}\' is not allowed as an argument. '\
+                    'Only these keys are allowed: {}'.format(key,
+                                                             cls.valid_fields)
+                raise ValueError(message)
+        for key in cls.required_fields:
+            if key not in kwargs.keys():
+                message = 'There must be an \'{}\' item in the input dict'\
+                    .format(key)
+                raise ValueError(message)
+        args = [kwargs.pop(key) for key in cls.required_fields]
+        return cls(*args)
+
+    @property
+    def uri(self):
+        """Return the URI of the object"""
+        return self.content['uri']
+
+    @property
+    # pylint: disable=no-self-use
+    def didl_metadata(self):
+        """Return the DIDL metadata for a URI
+
+        It is not yet know how to structure the metadata for URIs, so at the
+        moment the metadata consist purely of an empty DIDL-Lite container:
+
+        .. code :: xml
+
+         <DIDL-Lite/>
+
+        """
+        xml = XML.Element('DIDL-Lite')
+        return xml
+
+
 ###############################################################################
 # QUEUE                                                                       #
 ###############################################################################
@@ -745,13 +832,15 @@ class QueueItem(MusicInfoItem):
     @property
     def to_dict(self):
         """Get the dict representation of the instance."""
-        return self.content
+        return self.content.copy()
 
     @property
+    # pylint: disable=no-self-use
     def didl_metadata(self):
-        """Produce the DIDL metadata XML. CURRENTLY DISABLED."""
-        message = 'Queueitems cannot not yet form didl_metadata'
-        raise NotImplementedError(message)
+        """Produce the DIDL metadata XML."""
+        message = 'Queueitems are not meant to be re-added to the queue and '\
+            'therefore cannot create their own didl_metadata'
+        raise CannotCreateDIDLMetadata(message)
 
     @property
     def title(self):
@@ -834,18 +923,25 @@ class MusicServiceItem(MusicInfoItem):
         self.content = kwargs
 
     @classmethod
-    def from_xml(cls, xml, service):
+    def from_xml(cls, xml, service, parent_id):
         """Return a Music Service item generated from xml
 
         :param xml: Object XML. All items containing text are added to the
             content of the item. The class variable ``valid_fields`` of each of
             the classes list the valid fields (after translating the camel
-            case to underscore notation). Required fields are liste in the
+            case to underscore notation). Required fields are listed in the
             class variable by that name (where 'id' has been renamed to
             'item_id').
         :type xml: :py:class:`xml.etree.ElementTree.Element`
-        :param service: The music service instance that retrieved the element
-        :type service: :class:`soco.services.MusicServices`
+        :param service: The music service (plugin) instance that retrieved the
+            element. This service must contain ``id_to_extended_id`` and
+            ``form_uri`` methods and ``description`` and ``service_id``
+            attributes.
+        :type service: Instance of sub-class of
+            :class:`soco.plugins.SoCoPlugin`
+        :param parent_id: The parent ID of the item, will either be the
+            extended ID of another MusicServiceItem or of a search
+        :type parent_id: str
 
         For a track the XML can e.g. be on the following form:
 
@@ -876,19 +972,20 @@ class MusicServiceItem(MusicInfoItem):
            </trackMetadata>
          </mediaMetadata>
         """
+        # Add a few extra pieces of information
+        content = {'description': service.description,
+                   'service_id': service.service_id,
+                   'parent_id': parent_id}
         # Extract values from the XML
         all_text_elements = tags_with_text(xml)
-        content = {'username': service.username,
-                   'service_id': service.session_id,
-                   'description': service.description}
         for item in all_text_elements:
             tag = item.tag[len(NS['ms']) + 2:]  # Strip namespace
             tag = camel_to_underscore(tag)  # Convert to nice names
             if tag not in cls.valid_fields:
-                message = 'The info tag \'{}\' is not allowed for this item'.\
+                message = 'The info tag \'{0}\' is not allowed for this item'.\
                     format(tag)
                 raise ValueError(message)
-            content[camel_to_underscore(tag)] = item.text
+            content[tag] = item.text
 
         # Convert values for known types
         for key, value in content.items():
@@ -899,18 +996,30 @@ class MusicServiceItem(MusicInfoItem):
                 content[key] = True if value == 'true' else False
         # Rename a single item
         content['item_id'] = content.pop('id')
+        # And get the extended id
+        content['extended_id'] = service.id_to_extended_id(content['item_id'],
+                                                           cls)
+        # Add URI if there is one for the relevant class
+        uri = service.form_uri(content, cls)
+        if uri:
+            content['uri'] = uri
 
         # Check for all required values
         for key in cls.required_fields:
             if key not in content:
-                message = 'An XML field that correspond to the key \'{}\' is '\
-                    'required. See the docstring for help.'.format(key)
+                message = 'An XML field that correspond to the key \'{0}\' '\
+                    'is required. See the docstring for help.'.format(key)
 
         return cls.from_dict(content)
 
     @classmethod
     def from_dict(cls, dict_in):
-        """Initialize the class from a dict"""
+        """Initialize the class from a dict
+
+        :param dict_in: The dictionary that contains the item content. Required
+            fields are listed class variable by that name
+        :type dict_in: dict
+        """
         kwargs = dict_in.copy()
         args = [kwargs.pop(key) for key in cls.required_fields]
         return cls(*args, **kwargs)
@@ -920,51 +1029,6 @@ class MusicServiceItem(MusicInfoItem):
         """Return a copy of the content dict"""
         return self.content.copy()
 
-    @property
-    def item_id(self):
-        """Return the item id"""
-        return self.content['item_id']
-
-    @property
-    def title(self):
-        """Return the title"""
-        return self.content['title']
-
-    @property
-    def username(self):
-        """Return the username"""
-        return self.content['username']
-
-    @property
-    def service_id(self):
-        """Return the service id"""
-        return self.content['service_id']
-
-
-class MSTrack(MusicServiceItem):
-    """Music service track"""
-
-    item_class = 'object.item.audioItem.musicTrack'
-    valid_fields = [
-        'album', 'can_add_to_favorites', 'artist', 'album_artist_id', 'title',
-        'album_id', 'album_art_uri', 'album_artist', 'composer_id',
-        'item_type', 'composer', 'duration', 'can_skip', 'artist_id',
-        'can_play', 'id', 'mime_type', 'description'
-    ]
-    required_fields = [
-        'title', 'item_id', 'album_id', 'mime_type', 'description',
-        'service_id'
-    ]
-
-    def __init__(self, title, item_id, album_id, mime_type, description,
-                 service_id, **kwargs):
-        """Initialize MSTrack item"""
-        content = {'title': title, 'item_id': item_id, 'album_id': album_id,
-                   'mime_type': mime_type, 'description': description,
-                   'service_id': service_id}
-        content.update(kwargs)
-        super(MSTrack, self).__init__(**content)
-
     @property
     def didl_metadata(self):
         """Return the DIDL metadata for a Music Service Track
@@ -977,20 +1041,38 @@ class MSTrack(MusicServiceItem):
               xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
               xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
               xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
-           <item id="00030020...self.item_id..."
-              parentID="0004002c...self.album_id"
+           <item id="...self.extended_id..."
+              parentID="...self.parent_id..."
               restricted="true">
              <dc:title>...self.title...</dc:title>
              <upnp:class>...self.item_class...</upnp:class>
              <desc id="cdudn"
                 nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">
-               SA_RINCON5127_...self.username...
+               self.content['description']
              </desc>
            </item>
          </DIDL-Lite>
         """
-        # Main element, ugly yes, but I have given up on using namespaces with
-        # xml.etree.ElementTree
+        # Check if this item is meant to be played
+        if not self.can_play:
+            message = 'This item is not meant to be played and therefore '\
+                'also not to create its own didl_metadata'
+            raise CannotCreateDIDLMetadata(message)
+        # Check if we have the attributes to create the didl metadata:
+        for key in ['extended_id', 'title', 'item_class']:
+            if not hasattr(self, key):
+                message = 'The property \'{0}\' is not present on this item. '\
+                    'This indicates that this item was not meant to create '\
+                    'didl_metadata'.format(key)
+                raise CannotCreateDIDLMetadata(message)
+        if 'description' not in self.content:
+            message = 'The item for \'description\' is not present in '\
+                'self.content. This indicates that this item was not meant '\
+                'to create didl_metadata'
+            raise CannotCreateDIDLMetadata(message)
+
+        # Main element, ugly? yes! but I have given up on using namespaces
+        # with xml.etree.ElementTree
         item_attrib = {
             'xmlns:dc': 'http://purl.org/dc/elements/1.1/',
             'xmlns:upnp': 'urn:schemas-upnp-org:metadata-1-0/upnp/',
@@ -1000,11 +1082,15 @@ class MSTrack(MusicServiceItem):
         xml = XML.Element('DIDL-Lite', item_attrib)
         # Item sub element
         item_attrib = {
-            'parentID': '0004002c{}'.format(self.album_id),
+            'parentID': '',
             'restricted': 'true',
-            'id': '00030020{}'.format(self.item_id)
+            'id': self.extended_id
         }
+        # Only add the parent_id if we have it
+        if self.parent_id:
+            item_attrib['parentID'] = self.parent_id
         item = XML.SubElement(xml, 'item', item_attrib)
+
         # Add title and class
         XML.SubElement(item, 'dc:title').text = self.title
         XML.SubElement(item, 'upnp:class').text = self.item_class
@@ -1019,24 +1105,85 @@ class MSTrack(MusicServiceItem):
         return xml
 
     @property
-    def album_id(self):
-        """Return the album id"""
-        return self.content['album_id']
+    def item_id(self):
+        """Return the item id"""
+        return self.content['item_id']
 
     @property
-    def mime_type(self):
-        """Return the mime type"""
-        return self.content['mime_type']
+    def extended_id(self):
+        """Return the extended id"""
+        return self.content['extended_id']
+
+    @property
+    def title(self):
+        """Return the title"""
+        return self.content['title']
+
+    @property
+    def service_id(self):
+        """Return the service ID"""
+        return self.content['service_id']
+
+    @property
+    def can_play(self):
+        """Return a boolean for whether the item can be played"""
+        return bool(self.content.get('can_play'))
+
+    @property
+    def parent_id(self):
+        """Return the extended parent_id, if set, otherwise return None"""
+        return self.content.get('parent_id')
+
+    @property
+    def album_art_uri(self):
+        """Return the album art URI if set, otherwise return None"""
+        return self.content.get('album_art_uri')
+
+
+class MSTrack(MusicServiceItem):
+    """Class that represents a music service track"""
+
+    item_class = 'object.item.audioItem.musicTrack'
+    valid_fields = [
+        'album', 'can_add_to_favorites', 'artist', 'album_artist_id', 'title',
+        'album_id', 'album_art_uri', 'album_artist', 'composer_id',
+        'item_type', 'composer', 'duration', 'can_skip', 'artist_id',
+        'can_play', 'id', 'mime_type', 'description'
+    ]
+    # IMPORTANT. Keep this list, __init__ args and content in __init__ in sync
+    required_fields = ['title', 'item_id', 'extended_id', 'uri', 'description',
+                       'service_id']
+
+    def __init__(self, title, item_id, extended_id, uri, description,
+                 service_id, **kwargs):
+        """Initialize MSTrack item"""
+        content = {
+            'title': title, 'item_id': item_id, 'extended_id': extended_id,
+            'uri': uri, 'description': description, 'service_id': service_id,
+        }
+        content.update(kwargs)
+        super(MSTrack, self).__init__(**content)
+
+    @property
+    def album(self):
+        """Return the album title if set, otherwise return None"""
+        return self.content.get('album')
+
+    @property
+    def artist(self):
+        """Return the artist if set, otherwise return None"""
+        return self.content.get('artist')
+
+    @property
+    def duration(self):
+        """Return the duration if set, otherwise return None"""
+        return self.content.get('duration')
 
     @property
     def uri(self):
-        """Return the uri"""
+        """Return the URI"""
         # x-sonos-http:trackid_19356232.mp4?sid=20&amp;flags=32
-        return 'x-sonos-http:{}.{}?sid={}&flags=32'.format(
-            self.item_id,
-            MIME_TYPE_TO_EXTENSION[self.mime_type],
-            self.service_id
-        )
+        return self.content['uri']
 
 
 class MSAlbum(MusicServiceItem):
@@ -1045,188 +1192,182 @@ class MSAlbum(MusicServiceItem):
     item_class = 'object.container.album.musicAlbum'
     valid_fields = [
         'username', 'can_add_to_favorites', 'artist', 'title', 'album_art_uri',
-        'can_play', 'item_type', 'service_id', 'id', 'description'
+        'can_play', 'item_type', 'service_id', 'id', 'description',
+        'can_cache', 'artist_id', 'can_skip'
     ]
-    required_fields = ['title', 'item_id', 'description', 'service_id']
+    # IMPORTANT. Keep this list, __init__ args and content in __init__ in sync
+    required_fields = ['title', 'item_id', 'extended_id', 'uri', 'description',
+                       'service_id']
 
-    def __init__(self, title, item_id, description, service_id, **kwargs):
-        content = {'title': title, 'item_id': item_id,
-                   'description': description, 'service_id': service_id}
+    def __init__(self, title, item_id, extended_id, uri, description,
+                 service_id, **kwargs):
+        content = {
+            'title': title, 'item_id': item_id, 'extended_id': extended_id,
+            'uri': uri, 'description': description, 'service_id': service_id,
+        }
         content.update(kwargs)
         super(MSAlbum, self).__init__(**content)
 
     @property
-    def didl_metadata(self):
-        """Return the DIDL metadata for a Music Service Album
-
-        The metadata is on the form:
-
-        .. code :: xml
-
-         <DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/"
-              xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
-              xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
-              xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
-           <item id="0004002calbumid_22757081"
-                parentID="000d0064artistmainalbumsid_3694795"
-                restricted="true">
-             <dc:title>Aventine</dc:title>
-             <upnp:class>object.container.album.musicAlbum</upnp:class>
-             <desc id="cdudn"
-                  nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">
-               SA_RINCON5127_...self.username...
-             </desc>
-           </item>
-         </DIDL-Lite>
-
-        """
-        # Main element, ugly yes, but I have given up on using namespaces with
-        # xml.etree.ElementTree
-        item_attrib = {
-            'xmlns:dc': 'http://purl.org/dc/elements/1.1/',
-            'xmlns:upnp': 'urn:schemas-upnp-org:metadata-1-0/upnp/',
-            'xmlns:r': 'urn:schemas-rinconnetworks-com:metadata-1-0/',
-            'xmlns': 'urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/'
-        }
-        xml = XML.Element('DIDL-Lite', item_attrib)
-        # Item sub element
-        item_attrib = {
-            'parentID': '',
-            'restricted': 'true',
-            'id': '0004002c{}'.format(self.item_id)
-        }
-        item = XML.SubElement(xml, 'item', item_attrib)
-        # Add title and class
-        XML.SubElement(item, 'dc:title').text = self.title
-        XML.SubElement(item, 'upnp:class').text = self.item_class
-        # Add the desc element
-        desc_attrib = {
-            'id': 'cdudn',
-            'nameSpace': 'urn:schemas-rinconnetworks-com:metadata-1-0/'
-        }
-        desc = XML.SubElement(item, 'desc', desc_attrib)
-        desc.text = self.content['description']
-
-        return xml
+    def artist(self):
+        """Return the artist if set, otherwise return None"""
+        return self.content.get('artist')
 
     @property
     def uri(self):
-        """Return the uri"""
+        """Return the URI"""
         # x-rincon-cpcontainer:0004002calbumid_22757081
-        return 'x-rincon-cpcontainer:0004002c{}'.format(self.item_id)
+        return self.content['uri']
 
 
-class MSArtist(MusicServiceItem):
-    """Class that represents a music service artist"""
-
-    valid_fields = [
-        'username', 'can_add_to_favorites', 'artist', 'title', 'album_art_uri',
-        'item_type', 'id', 'service_id', 'description'
-    ]
-    # Since MSArtist cannot produce didl_metadata, they are strictly required,
-    # but it makes sense to require them anyway, since that are the fields that
-    # that describe the item
-    required_fields = ['title', 'item_id', 'description', 'service_id']
-
-    def __init__(self, title, item_id, description, service_id, **kwargs):
-        content = {'title': title, 'item_id': item_id,
-                   'description': description, 'service_id': service_id}
-        content.update(kwargs)
-        super(MSArtist, self).__init__(**content)
-
-
-class MSPlaylist(MusicServiceItem):
-    """Class that represents a Music Service Playlist"""
+class MSAlbumList(MusicServiceItem):
+    """Class that represents a Music Service Album List"""
 
     item_class = 'object.container.albumlist'
     valid_fields = [
-        'username', 'can_add_to_favorites', 'artist', 'title', 'album_art_uri',
-        'can_play', 'item_type', 'id', 'service_id', 'artist_id',
-        'can_enumerate', 'description'
+        'id', 'title', 'item_type', 'artist', 'artist_id', 'can_play',
+        'can_enumerate', 'can_add_to_favorites', 'album_art_uri', 'can_cache'
     ]
-    required_fields = ['title', 'item_id', 'description', 'service_id']
+    # IMPORTANT. Keep this list, __init__ args and content in __init__ in sync
+    required_fields = ['title', 'item_id', 'extended_id', 'uri', 'description',
+                       'service_id']
 
-    def __init__(self, title, item_id, description, service_id, **kwargs):
-        content = {'title': title, 'item_id': item_id,
-                   'description': description, 'service_id': service_id}
+    def __init__(self, title, item_id, extended_id, uri, description,
+                 service_id, **kwargs):
+        content = {
+            'title': title, 'item_id': item_id, 'extended_id': extended_id,
+            'uri': uri, 'description': description, 'service_id': service_id,
+        }
+        content.update(kwargs)
+        super(MSAlbumList, self).__init__(**content)
+
+    @property
+    def uri(self):
+        """Return the URI"""
+        # x-rincon-cpcontainer:000d006cplaylistid_26b18dbb-fd35-40bd-8d4f-
+        # 8669bfc9f712
+        return self.content['uri']
+
+
+class MSPlaylist(MusicServiceItem):
+    """Class that represents a Music Service Play List"""
+
+    item_class = 'object.container.albumlist'
+    valid_fields = ['id', 'item_type', 'title', 'can_play', 'can_cache',
+                    'album_art_uri', 'artist', 'can_enumerate',
+                    'can_add_to_favorites', 'artist_id']
+    # IMPORTANT. Keep this list, __init__ args and content in __init__ in sync
+    required_fields = ['title', 'item_id', 'extended_id', 'uri', 'description',
+                       'service_id']
+
+    def __init__(self, title, item_id, extended_id, uri, description,
+                 service_id, **kwargs):
+        content = {
+            'title': title, 'item_id': item_id, 'extended_id': extended_id,
+            'uri': uri, 'description': description, 'service_id': service_id,
+        }
         content.update(kwargs)
         super(MSPlaylist, self).__init__(**content)
 
     @property
-    def didl_metadata(self):
-        """Return the DIDL metadata for a Music Service Playlist
-
-        The metadata is on the form:
-
-        .. code :: xml
-
-         <DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/"
-              xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
-              xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
-              xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
-           <item id="000d006cplaylist_bf148cdf-a69b-4da1-b3a3-d94315abe9a4"
-                parentID="00020064playlistsearch:a"
-                restricted="true">
-             <dc:title>WiMP: &#197;rets mest spillede sange</dc:title>
-             <upnp:class>object.container.albumlist</upnp:class>
-             <desc id="cdudn"
-                   nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">
-               SA_RINCON5127_...self.username...
-             </desc>
-           </item>
-         </DIDL-Lite>
+    def uri(self):
+        """Return the URI"""
+        # x-rincon-cpcontainer:000d006cplaylistid_c86ddf26-8ec5-483e-b292-
+        # abe18848e89e
+        return self.content['uri']
 
 
-        """
-        # Main element, ugly yes, but I have given up on using namespaces with
-        # xml.etree.ElementTree
-        item_attrib = {
-            'xmlns:dc': 'http://purl.org/dc/elements/1.1/',
-            'xmlns:upnp': 'urn:schemas-upnp-org:metadata-1-0/upnp/',
-            'xmlns:r': 'urn:schemas-rinconnetworks-com:metadata-1-0/',
-            'xmlns': 'urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/'
+class MSArtistTracklist(MusicServiceItem):
+    """Class that represents a Music Service Artist Track List"""
+
+    item_class = 'object.container.playlistContainer.sameArtist'
+    valid_fields = ['id', 'title', 'item_type', 'can_play', 'album_art_uri']
+    # IMPORTANT. Keep this list, __init__ args and content in __init__ in sync
+    required_fields = ['title', 'item_id', 'extended_id', 'uri', 'description',
+                       'service_id']
+
+    def __init__(self, title, item_id, extended_id, uri, description,
+                 service_id, **kwargs):
+        content = {
+            'title': title, 'item_id': item_id, 'extended_id': extended_id,
+            'uri': uri, 'description': description, 'service_id': service_id,
         }
-        xml = XML.Element('DIDL-Lite', item_attrib)
-        # Item sub element
-        item_attrib = {
-            'parentID': '',
-            'restricted': 'true',
-            'id': '000d006c{}'.format(self.item_id)
-        }
-        item = XML.SubElement(xml, 'item', item_attrib)
-        # Add title and class
-        XML.SubElement(item, 'dc:title').text = self.title
-        XML.SubElement(item, 'upnp:class').text = self.item_class
-        # Add the desc element
-        desc_attrib = {
-            'id': 'cdudn',
-            'nameSpace': 'urn:schemas-rinconnetworks-com:metadata-1-0/'
-        }
-        desc = XML.SubElement(item, 'desc', desc_attrib)
-        desc.text = self.content['description']
-
-        return xml
+        content.update(kwargs)
+        super(MSArtistTracklist, self).__init__(**content)
 
     @property
     def uri(self):
-        """Return the uri"""
-        # x-rincon-cpcontainer:000d006cplaylist_bf148cdf-a69b-4da1-b3a3-
-        # d94315abe9a4
-        return 'x-rincon-cpcontainer:000d006c{}'.format(self.item_id)
+        """Return the URI"""
+        # x-rincon-cpcontainer:100f006cartistpopsongsid_1566
+        return 'x-rincon-cpcontainer:100f006c{0}'.format(self.item_id)
+
+
+class MSArtist(MusicServiceItem):
+    """Class that represents a Music Service Artist"""
+
+    valid_fields = [
+        'username', 'can_add_to_favorites', 'artist', 'title', 'album_art_uri',
+        'item_type', 'id', 'service_id', 'description', 'can_cache'
+    ]
+    # Since MSArtist cannot produce didl_metadata, they are not strictly
+    # required, but it makes sense to require them anyway, since they are the
+    # fields that that describe the item
+    # IMPORTANT. Keep this list, __init__ args and content in __init__ in sync
+    required_fields = ['title', 'item_id', 'extended_id', 'service_id']
+
+    def __init__(self, title, item_id, extended_id, service_id, **kwargs):
+        content = {'title': title, 'item_id': item_id,
+                   'extended_id': extended_id, 'service_id': service_id}
+        content.update(kwargs)
+        super(MSArtist, self).__init__(**content)
+
+
+class MSFavorites(MusicServiceItem):
+    """Class that represents a Music Service Favorite"""
+
+    valid_fields = ['id', 'item_type', 'title', 'can_play', 'can_cache',
+                    'album_art_uri']
+    # Since MSFavorites cannot produce didl_metadata, they are not strictly
+    # required, but it makes sense to require them anyway, since they are the
+    # fields that that describe the item
+    # IMPORTANT. Keep this list, __init__ args and content in __init__ in sync
+    required_fields = ['title', 'item_id', 'extended_id', 'service_id']
+
+    def __init__(self, title, item_id, extended_id, service_id, **kwargs):
+        content = {'title': title, 'item_id': item_id,
+                   'extended_id': extended_id, 'service_id': service_id}
+        content.update(kwargs)
+        super(MSFavorites, self).__init__(**content)
+
+
+class MSCollection(MusicServiceItem):
+    """Class that represents a Music Service Collection"""
+
+    valid_fields = ['id', 'item_type', 'title', 'can_play', 'can_cache',
+                    'album_art_uri']
+    # Since MSCollection cannot produce didl_metadata, they are not strictly
+    # required, but it makes sense to require them anyway, since they are the
+    # fields that that describe the item
+    # IMPORTANT. Keep this list, __init__ args and content in __init__ in sync
+    required_fields = ['title', 'item_id', 'extended_id', 'service_id']
+
+    def __init__(self, title, item_id, extended_id, service_id, **kwargs):
+        content = {'title': title, 'item_id': item_id,
+                   'extended_id': extended_id, 'service_id': service_id}
+        content.update(kwargs)
+        super(MSCollection, self).__init__(**content)
 
 
 PARENT_ID_TO_CLASS = {'A:TRACKS': MLTrack, 'A:ALBUM': MLAlbum,
                       'A:ARTIST': MLArtist, 'A:ALBUMARTIST': MLAlbumArtist,
                       'A:GENRE': MLGenre, 'A:COMPOSER': MLComposer,
-                      'A:PLAYLISTS': MLPlaylist, 'S:': MLShare}
+                      'A:PLAYLISTS': MLPlaylist, 'S:': MLShare,
+                      'SQ:': MLSonosPlaylist}
 
 MS_TYPE_TO_CLASS = {'artist': MSArtist, 'album': MSAlbum, 'track': MSTrack,
-                    'albumList': MSPlaylist}
-
-
-MIME_TYPE_TO_EXTENSION = {
-    'audio/aac': 'mp4'
-}
+                    'albumList': MSAlbumList, 'favorites': MSFavorites,
+                    'collection': MSCollection, 'playlist': MSPlaylist,
+                    'artistTrackList': MSArtistTracklist}
 
 NS = {
     'dc': 'http://purl.org/dc/elements/1.1/',
